//---------------------------------------------------------------------------

#ifndef uFunctionDelegateHPP
#define uFunctionDelegateHPP
//---------------------------------------------------------------------------
#include <System.JSON.hpp>
#include <System.Classes.hpp>

template <class fnT>
String __fastcall fnGetFnType(fnT fn)
{
  return Format("%s", ARRAYOFCONST((typeid(fn).name())));
}

template <class fnT>
String __fastcall fnGetFnTypeReturn(fnT fn)
{
  String s = fnGetFnType(fn);
  return s.SubString(1, s.Pos(" ") - 1);
}

template <class fnT>
String __fastcall fnGetFnTypeCall(fnT fn)
{
  String s = fnGetFnType(fn);
  int offx = s.Pos("(") + 1;
  return s.SubString(offx, s.Pos(")") - offx);
}

template <class fnT>
String __fastcall fnGetFnTypeParms(fnT fn)
{
  String s = fnGetFnType(fn);
  int offx = s.LastDelimiter("(") + 1;
  return s.SubString(offx, s.LastDelimiter(")") - offx);
}

template <class fnT>
TStringList * __fastcall fnNewFnTypeParmList(fnT fn)
{
  TStringList *Result = new TStringList();
  Result->CommaText = fnGetFnTypeParms(fn);
  return Result;
}

template <class T>
TJSONObject * __fastcall fnGetTypeVar(T aVar)
{
  TJSONObject *Result = NULL;
  PTypeInfo Info = __delphirtti(T);
  String Kind, Name;

  if (Info != NULL) {
    Kind = GetEnumName(__delphirtti(TTypeKind), Info->Kind);
    Result = new TJSONObject();

    switch (Info->Kind) {
    case tkString: case tkLString: case tkUString: case tkWString:
    case tkChar: case tkWChar:
      Result->AddPair(new TJSONPair(Kind, Format("%s", ARRAYOFCONST((aVar)))));
      break;

    case tkInteger: case tkInt64:
      Result->AddPair(new TJSONPair(Kind, Format("%d", ARRAYOFCONST((aVar)))));
      break;

    case tkFloat:
      Result->AddPair(new TJSONPair(Kind, Format("%f", ARRAYOFCONST((aVar)))));
      break;

    case tkUnknown:
      Result->AddPair(new TJSONPair(typeid(aVar).name(), Format("%s", ARRAYOFCONST((aVar)))));
      break;

    default:;
    // TODO: consider whether or not to use typeid(aVar).name() ?
    }
  }
  return Result;
}

template <typename T0=void, typename T1=void, typename T2=void, typename T3=void, typename T4=void>
class TjsParm : public System::TObject
{
private:
  TJSONObject *m_Parms;
  TJSONArray *m_Array;

private:
  int __fastcall GetParmCount()
  {
    TJSONArray *jsArr = dynamic_cast<TJSONArray *>(m_Parms->Pairs[0]->JsonValue);
    return (jsArr == NULL) ? 0: (jsArr->Count);
  }

  TJSONObject * __fastcall GetParms() const
  {
    return m_Parms;
  }

  String __fastcall GetTypeKindString(int index)
  {
    String Result;
    TJSONArray *jsArr;
    TJSONObject *jsObj;
    jsArr = dynamic_cast<TJSONArray *>(m_Parms->Pairs[0]->JsonValue);

    if ((jsArr != NULL) && (jsArr->Count > 0)) {
      jsObj = dynamic_cast<TJSONObject *>(jsArr->Items[index]);
      if (jsObj != NULL) {
        Result = jsObj->Pairs[0]->JsonString->Value();
      }
    }
    return Result;
  }

  TTypeKind __fastcall GetTypeKind(int index)
  {
    // If the conversion string is not within the definition, which of return value is 255
    return GetEnumValue(__delphirtti(TTypeKind), GetTypeKindString(index));
  }

  TValue __fastcall GetTValue(int index)
  {
    TValue Result;
    TJSONArray *jsArr;
    TJSONObject *jsObj;
    jsArr = dynamic_cast<TJSONArray *>(m_Parms->Pairs[0]->JsonValue);
    TTypeKind type =  GetTypeKind(index);
    String Val, name;

    if ((jsArr != NULL) && (jsArr->Count > 0)) {
      jsObj = dynamic_cast<TJSONObject *>(jsArr->Items[index]);
      if (jsObj != NULL) {
        Val = jsObj->Pairs[0]->JsonValue->Value();
        switch (type) {
        case tkString: case tkLString: case tkUString: case tkWString:
        case tkChar: case tkWChar:
          Result = TValue::From<String>(Val);
          break;

        case tkInteger:
          Result = TValue::From<int>(StrToInt((Val)));
          break;

        case tkInt64:
          Result = TValue::From<__int64>(StrToInt64((Val)));
          break;

        case tkFloat:
          Result = TValue::From<float>(StrToFloat((Val)));
          break;

        case 255:
          name = GetTypeKindString(index);
          if (name == L"const char *" || name == L"const wchar_t *") {
            Result = TValue::From<String>(Val);
          }
          else {
          // TODO:
          }
          break;

        default:;
        }
      }
    }
    return Result;
  }

  void __fastcall Add(T0 Var0)
  {
    m_Array->AddElement(fnGetTypeVar<T0>(Var0));
  }

  void __fastcall Add(T0 Var0, T1 Var1)
  {
    Add(Var0);
    m_Array->AddElement(fnGetTypeVar<T1>(Var1));
  }

  void __fastcall Add(T0 Var0, T1 Var1, T2 Var2)
  {
    Add(Var0, Var1);
    m_Array->AddElement(fnGetTypeVar<T2>(Var2));
  }

  void __fastcall Add(T0 Var0, T1 Var1, T1 Var2, T3 Var3)
  {
    Add(Var0, Var1, Var2);
    m_Array->AddElement(fnGetTypeVar<T3>(Var3));
  }

  void __fastcall Add(T0 Var0, T1 Var1, T2 Var2, T3 Var3, T4 Var4)
  {
    Add(Var0, Var1, Var2, Var3);
    m_Array->AddElement(fnGetTypeVar<T4>(Var4));
  }

  void __fastcall CreateJs()
  {
    m_Parms = new TJSONObject();
    m_Array = new TJSONArray();
    m_Parms->AddPair("Parameter", m_Array);
  }

public:
  __fastcall TjsParm() : System::TObject()
  {
    CreateJs();
  }

  __fastcall TjsParm(T0 Var0) : System::TObject()
  {
    CreateJs();
    Add(Var0);
  }

  __fastcall TjsParm(T0 Var0, T1 Var1) : System::TObject()
  {
    CreateJs();
    Add(Var0, Var1);
  }

  __fastcall TjsParm(T0 Var0, T1 Var1, T2 Var2) : System::TObject()
  {
    CreateJs();
    Add(Var0, Var1, Var2);
  }

  __fastcall TjsParm(T0 Var0, T1 Var1, T2 Var2, T3 Var3) : System::TObject()
  {
    CreateJs();
    Add(Var0, Var1, Var2, Var3);
  }

  __fastcall TjsParm(T0 Var0, T1 Var1, T2 Var2, T3 Var3, T4 Var4) : System::TObject()
  {
    CreateJs();
    Add(Var0, Var1, Var2, Var3, Var4);
  }

  __fastcall TjsParm(TJSONValue *aVar) : System::TObject()
  {
    TJSONValue *Var;
    if (aVar == NULL) {
      CreateJs();
    }
    else {
      Var = TJSONObject::ParseJSONValue(aVar->ToJSON());
      m_Parms = dynamic_cast<TJSONObject *>(Var);
    }
  }

  __fastcall ~TjsParm()
  {
    delete m_Parms;
  }

public:
  __property TValue Value[int index] = {read=GetTValue};
  __property String TypeKindString[int index] = {read=GetTypeKindString};
  __property int Count = {read=GetParmCount};
  __property TJSONObject const *Parms = {read=GetParms};

};

// the return event type which delegate's task has completed
typedef void __fastcall (*pfnDone)(String);
typedef void __fastcall (__closure *TDone)(String);
typedef void __fastcall (__closure *TDoneEvent)(TObject *Sender, String);

typedef String __fastcall (*Mypfn)(String);
// only for function type which is 'String __fastcall (*)(String)'
class TMyDelegate : public TThread
{
private:
  Mypfn m_fn;
  pfnDone m_fnDone;
  TDone m_Done;
  TDoneEvent m_DoneEvent;
  int m_SynMode;
  String m_Parm, m_Result;

  void __fastcall Init(Mypfn afn, String aParm, int aMode)
  {
    m_fnDone = NULL;
    m_Done = NULL;
    m_DoneEvent = NULL;
    m_fn = afn;
    m_Parm = aParm;
    m_SynMode = aMode;
    FreeOnTerminate = true;
  }

public:
  __fastcall ~TMyDelegate() {}
  __fastcall TMyDelegate(Mypfn afn, pfnDone aDone, String aParm, int aMode = 1) : TThread(false)
  {
    Init(afn, aParm, aMode);
    m_fnDone = aDone;
  }

  __fastcall TMyDelegate(Mypfn afn, TDone aDone, String aParm, int aMode = 1) : TThread(false)
  {
    Init(afn, aParm, aMode);
    m_Done = aDone;
  }

  __fastcall TMyDelegate(Mypfn afn, TDoneEvent aEvent, String aParm, int aMode = 1) : TThread(false)
  {
    Init(afn, aParm, aMode);
    m_DoneEvent = aEvent;
  }

protected:
  void __fastcall Done()
  {
    // return task result
    if (m_fnDone != NULL) {
      m_fnDone(m_Result);
    }
    else if (m_Done != NULL) {
      m_Done(m_Result);
    }
    else if (m_DoneEvent != NULL) {
      m_DoneEvent(this, m_Result);
    }
  }

  void __fastcall DoSyn()
  {
    // thread safe
    if (m_SynMode == 1) {
      Synchronize(Done);
    }
    // not thread safe
    else {
      Done();
    }
  }

  void __fastcall Execute(void)
  {
    m_Result = m_fn(m_Parm);
    DoSyn();
  }
};

// only for function type which are
// 'String __fastcall(*)()', 'String __fastcall(*)(String)',
// 'String __fastcall(*)(Parms0(String), Parms1(String), ...ParmsN(String))'
template <typename fnT, typename EventT>
class TThxDelegate : public TCppInterfacedObject<TThreadProcedure>
{
private:
  const static m_ParmsMaxCount = 8;
  String m_Parms[m_ParmsMaxCount];
  EventT m_DoneEvent;
  fnT m_fn;
  int m_AddCount;

  void __fastcall Init(fnT afn, EventT aEvent)
  {
    m_fn = afn;
    m_DoneEvent = aEvent;
    m_AddCount = 0;
  }

  void __fastcall AddParam(String aParam)
  {
    m_Parms[m_AddCount++] = aParam;
  }

public:
  INTFOBJECT_IMPL_IUNKNOWN(TInterfacedObject);
  __fastcall ~TThxDelegate() {}

  __fastcall TThxDelegate(fnT afn, EventT aEvent)
  {
    Init(afn, aEvent);
  }

  __fastcall TThxDelegate(fnT afn, EventT aEvent, String aParm0)
  {
    Init(afn, aEvent);
    AddParam(aParm0);
  }

  __fastcall TThxDelegate(fnT afn, EventT aEvent, String aParm0, String aParm1)
  {
    Init(afn, aEvent);
    AddParam(aParm0); AddParam(aParm1);
  }

  __fastcall TThxDelegate(fnT afn, EventT aEvent, String aParm0, String aParm1, String aParm2)
  {
    Init(afn, aEvent);
    AddParam(aParm0); AddParam(aParm1); AddParam(aParm2);
  }

  void __fastcall Invoke()
  {
    String Result;
    if (m_AddCount == 0) {
      typedef String __fastcall (*pfn)();
      pfn fn = (pfn)m_fn;
      Result = fn();
    }
    else if (m_AddCount == 1) {
      typedef String __fastcall (*pfn0)(String);
      pfn0 fn0 = (pfn0)m_fn;
      Result = fn0(m_Parms[0]);
    }
    else if (m_AddCount == 2) {
      typedef String __fastcall (*pfn1)(String, String);
      pfn1 fn1 = (pfn1)m_fn;
      Result = fn1(m_Parms[0], m_Parms[1]);
    }
    else if (m_AddCount == 3) {
      typedef String __fastcall (*pfn2)(String, String, String);
      pfn2 fn2 = (pfn2)m_fn;
      Result = fn2(m_Parms[0], m_Parms[1], m_Parms[2]);
    }

    if (m_DoneEvent != NULL) {
      m_DoneEvent(Result);
    }
  }
};

template <typename fnT, typename EventT>
class TfnDelegate : public TCppInterfacedObject<TProc>
{
protected:
  TValue m_Result;
  TjsParm<> *m_pm;
  EventT m_DoneEvent;
  fnT m_fn;
  int m_SynMode;

protected:
  void __fastcall Done()
  {
    // return task result
    if (m_DoneEvent != NULL) {
      m_DoneEvent(m_Result.ToString());
    }
  }

  void __fastcall DoSyn()
  {
    // thread safe
    if (m_SynMode == 1) {
      TThread::Synchronize(TThread::CurrentThread, Done);
    }
    // not thread safe
    else {
      Done();
    }
  }

public:
  INTFOBJECT_IMPL_IUNKNOWN(TInterfacedObject);

  __fastcall TfnDelegate(fnT afn, EventT aEvent, TJSONValue *aParm)
  {
    m_fn = afn;
    m_DoneEvent = aEvent;
    m_pm = new TjsParm<>(aParm);
    // thread safe
    m_SynMode = 1;
  }

  __fastcall ~TfnDelegate()
  {
    delete m_pm;
  }
};

#define Delegate_IMPL_Construct(ParmNum) \
class TfnDelegate##ParmNum : public TfnDelegate<fnT, EventT> \
{ \
public: \
  __fastcall TfnDelegate##ParmNum(fnT afn, EventT aEvent, TJSONValue *aParm) \
    : TfnDelegate<fnT, EventT>(afn, aEvent, aParm) {}

// procedure type which has none input parameters
template <typename fnT, typename EventT>
  Delegate_IMPL_Construct(TProc);

  void __fastcall Invoke()
  {
    m_fn();
    DoSyn();
  }

  static void Delegate(fnT fn, EventT aEvent)
  {
    procDelegate(fn, aEvent);
  }
};
template<typename fnT, typename EventT>
void __fastcall procDelegate(fnT fn, EventT aEvent)
{
  TThread::CreateAnonymousThread(_di_TProc(new TfnDelegateTProc<fnT, EventT>
      (fn, aEvent, NULL)))->Start();
}

// procedure type which has one input parameters
template <typename fnT, typename EventT, typename T0>
  Delegate_IMPL_Construct(TProc__1);

  void __fastcall Invoke()
  {
    m_fn(m_pm->Value[0].AsType<T0>());
    DoSyn();
  }

  static void Delegate(fnT fn, EventT aEvent, T0 t0)
  {
    procDelegate(fn, aEvent, t0);
  }
};
template<typename fnT, typename EventT, typename T0>
void __fastcall procDelegate(fnT fn, EventT aEvent, T0 t0)
{
  TjsParm<T0>* pm = new TjsParm<T0>(t0);
  TThread::CreateAnonymousThread(_di_TProc(new TfnDelegateTProc__1<fnT, EventT, T0>
      (fn, aEvent, pm->Parms)))->Start();
  delete pm;
}

// procedure type which has two input parameters
template <typename fnT, typename EventT, typename T0, typename T1>
  Delegate_IMPL_Construct(TProc__2);

  void __fastcall Invoke()
  {
    m_fn(m_pm->Value[0].AsType<T0>(), m_pm->Value[1].AsType<T1>());
    DoSyn();
  }

  static void Delegate(fnT fn, EventT aEvent, T0 t0, T1 t1)
  {
    procDelegate(fn, aEvent, t0, t1);
  }
};
template<typename fnT, typename EventT, typename T0, typename T1>
void __fastcall procDelegate(fnT fn, EventT aEvent, T0 t0, T1 t1)
{
  TjsParm<T0, T1>* pm = new TjsParm<T0, T1>(t0, t1);
  TThread::CreateAnonymousThread(_di_TProc(new TfnDelegateTProc__2<fnT, EventT, T0, T1>
      (fn, aEvent, pm->Parms)))->Start();
  delete pm;
}

// procedure type which has three input parameters
template <typename fnT, typename EventT, typename T0, typename T1, typename T2>
  Delegate_IMPL_Construct(TProc__3);

  void __fastcall Invoke()
  {
    m_fn(m_pm->Value[0].AsType<T0>(), m_pm->Value[1].AsType<T1>(), m_pm->Value[2].AsType<T2>());
    DoSyn();
  }

  static void Delegate(fnT fn, EventT aEvent, T0 t0, T1 t1, T2 t2)
  {
    procDelegate(fn, aEvent, t0, t1, t2);
  }
};
template<typename fnT, typename EventT, typename T0, typename T1, typename T2>
void __fastcall procDelegate(fnT fn, EventT aEvent, T0 t0, T1 t1, T2 t2)
{
  TjsParm<T0, T1, T2>* pm = new TjsParm<T0, T1, T2>(t0, t1, t2);
  TThread::CreateAnonymousThread(_di_TProc(new TfnDelegateTProc__3<fnT, EventT, T0, T1, T2>
      (fn, aEvent, pm->Parms)))->Start();
  delete pm;
}

// function type which has has none input parameters
template <typename fnT, typename EventT>
  Delegate_IMPL_Construct(TFunc);

  void __fastcall Invoke()
  {
    m_Result = TValue::From<String>(m_fn());
    DoSyn();
  }

  static void Delegate(fnT fn, EventT aEvent)
  {
    fnDelegate(fn, aEvent);
  }
};
template<typename fnT, typename EventT>
void __fastcall fnDelegate(fnT fn, EventT aEvent)
{
  TThread::CreateAnonymousThread(_di_TProc(new TfnDelegateTFunc<fnT, EventT>
      (fn, aEvent, NULL)))->Start();
}

// function type which has one input parameters
template <typename fnT, typename EventT, typename T0>
  Delegate_IMPL_Construct(TFunc__1);

  void __fastcall Invoke()
  {
    m_Result = TValue::From<String>(m_fn(m_pm->Value[0].AsType<T0>()));
    DoSyn();
  }

  static void Delegate(fnT fn, EventT aEvent, T0 t0)
  {
    fnDelegate(fn, aEvent, t0);
  }
};
template<typename fnT, typename EventT, typename T0>
void __fastcall fnDelegate(fnT fn, EventT aEvent, T0 t0)
{
  TjsParm<T0>* pm = new TjsParm<T0>(t0);
  TThread::CreateAnonymousThread(_di_TProc(new TfnDelegateTFunc__1<fnT, EventT, T0>
      (fn, aEvent, pm->Parms)))->Start();
  delete pm;
}

// function type which has two input parameters
template <typename fnT, typename EventT, typename T0, typename T1>
  Delegate_IMPL_Construct(TFunc__2);

  void __fastcall Invoke()
  {
    m_Result = TValue::From<String>(m_fn(m_pm->Value[0].AsType<T0>(), m_pm->Value[1].AsType<T1>()));
    DoSyn();
  }

  static void Delegate(fnT fn, EventT aEvent, T0 t0, T1 t1)
  {
    fnDelegate(fn, aEvent, t0, t1);
  }
};
template<typename fnT, typename EventT, typename T0, typename T1>
void __fastcall fnDelegate(fnT fn, EventT aEvent, T0 t0, T1 t1)
{
  TjsParm<T0, T1>* pm = new TjsParm<T0, T1>(t0, t1);
  TThread::CreateAnonymousThread(_di_TProc(new TfnDelegateTFunc__2<fnT, EventT, T0, T1>
      (fn, aEvent, pm->Parms)))->Start();
  delete pm;
}

// function type which has three input parameters
template <typename fnT, typename EventT, typename T0, typename T1, typename T2>
  Delegate_IMPL_Construct(TFunc__3);

  void __fastcall Invoke()
  {
    m_Result = TValue::From<String>(m_fn(m_pm->Value[0].AsType<T0>(), m_pm->Value[1].AsType<T1>(),
        m_pm->Value[2].AsType<T2>()));
    DoSyn();
  }

  static void Delegate(fnT fn, EventT aEvent, T0 t0, T1 t1, T2 t2)
  {
    fnDelegate(fn, aEvent, t0, t1, t2);
  }
};
template<typename fnT, typename EventT, typename T0, typename T1, typename T2>
void __fastcall fnDelegate(fnT fn, EventT aEvent, T0 t0, T1 t1, T2 t2)
{
  TjsParm<T0, T1, T2>* pm = new TjsParm<T0, T1, T2>(t0, t1, t2);
  TThread::CreateAnonymousThread(_di_TProc(new TfnDelegateTFunc__3<fnT, EventT, T0, T1, T2>
      (fn, aEvent, pm->Parms)))->Start();
  delete pm;
}





#endif
